//@ts-check
import { Plugin } from "prosemirror-state"
import { keymap } from "prosemirror-keymap"
import {
  buildInputRules,
  buildKeymap,
  buildMenuItems,
} from "prosemirror-example-setup";
import { baseKeymap } from "prosemirror-commands";
import { dropCursor } from "prosemirror-dropcursor"
import { gapCursor } from "prosemirror-gapcursor"
import {
  menuBar,
  MenuItem,
  Dropdown,
  wrapItem,
  undoItem as prosemirrorUndoItem,
  redoItem as prosemirrorRedoItem,
  liftItem as prosemirrorLiftItem,
} from "prosemirror-menu";
import { yUndoPluginKey } from "y-prosemirror";
/**@typedef {import('prosemirror-model').Schema} Schema */

/**
 * @typedef {{
 *   schema: Schema
 *   undo: import('y-prosemirror').undo,
 *   redo: import('y-prosemirror').redo,
 * }} Options
 */

/**
 * @param {Options} options
 */
export default function getProsemirrorSetup(options) {
  // A modified `prosemirror-example-setup/src/index.ts`.

  const menuItemsBase = buildMenuItems(options.schema);
  const menuItemsFinal = [
    [
      new Dropdown(
        [
          new MenuItem({
            run: options.undo,
            label: "Undo typing",
            title: prosemirrorUndoItem.spec.title,
            class: prosemirrorUndoItem.spec.class,
            css:   prosemirrorUndoItem.spec.css,
          }),
          new MenuItem({
            run: options.redo,
            label: "Redo typing",
            title: prosemirrorRedoItem.spec.title,
            class: prosemirrorRedoItem.spec.class,
            css:   prosemirrorRedoItem.spec.css,
          }),
          new MenuItem({
            run: toggleDark,
            class: 'theme-button',
            title: 'Toggle color theme',
            label: 'Toggle theme'
          }),
        ], {
          label: '☰',
          title: 'Menu',
          class: 'hamburger-menu-item',
        },
      ),
      new Dropdown(
        [
          menuItemsBase.makeParagraph,
          menuItemsBase.makeHead1,
          menuItemsBase.makeHead2,
          menuItemsBase.makeHead3,
          wrapItem(options.schema.nodes.bullet_list, {
            label: "Bullet list",
          }),
          wrapItem(options.schema.nodes.ordered_list, {
            label: "Ordered list",
          }),
          new MenuItem({
            run:    prosemirrorLiftItem.spec.run,
            label:  prosemirrorLiftItem.spec.title,
            select: prosemirrorLiftItem.spec.select,
          }),
          wrapItem(options.schema.nodes.blockquote, {
            label: "Block quote",
          }),
          new MenuItem({
            ...menuItemsBase.insertHorizontalRule.spec,
            label: "Insert line",
          }),
        ], {
          label: 'Paragraph',
        },
      ),
      // menuItems.makeParagraph,
      menuItemsBase.toggleStrong,
      menuItemsBase.toggleEm,
      menuItemsBase.toggleCode,
    ],
  ];

  const plugins = [
    buildInputRules(options.schema),
    keymap(buildKeymap(options.schema, undefined)),
    keymap(baseKeymap),
    dropCursor(),
    gapCursor(),
    menuBar({
      // TODO `floating` doesn't work the way we want it to on iOS. See
      // https://codeberg.org/webxdc/editor/issues/11
      // Also maybe just replace it with `position: sticky`, like here:
      // https://codeberg.org/webxdc/editor/pulls/32
      floating: true,
      // content: menuItems.fullMenu,
      content: menuItemsFinal,
    }),
  ]
  return plugins.concat(new Plugin({
    props: {
      attributes: {class: "ProseMirror-example-setup-style"}
    }
  }))
}

function toggleDark() {
  const agentPrefersDark = window.matchMedia
    ?.('(prefers-color-scheme: dark)').matches
  const newWantedThemeIsDark =
    !document.documentElement.classList.contains('dark');
  if (agentPrefersDark === newWantedThemeIsDark) {
    // System theme and wanted theme match, let's just keep it that way.
    localStorage.removeItem('color-schema');
  } else {
    localStorage.setItem(
      'color-schema',
      newWantedThemeIsDark ? 'dark' : 'light'
    );
  }
  document.documentElement.classList.toggle('dark');
}

// Looked up in `y-prosemirror/src/plugins/undo-plugin.js
function getYjsUndoManager(state) {
  const undoManager = /**@type {import('yjs').UndoManager | undefined}*/(
    yUndoPluginKey.getState(state).undoManager
  );
  return undoManager;
}
